// hpstat: High-performance statistics implementations
// Copyright © 2023  Lee Yingtong Li (RunasSudo)
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

pub struct AndersonBjorckRootFinder {
	bound_lower: f64,
	bound_upper: f64,
	value_lower: f64,
	value_upper: f64,
	last_sign: f64  // Sign of the function at last evaluation (1.0 or -1.0) or 0.0 if first iteration
}

impl AndersonBjorckRootFinder {
	pub fn new(bound_lower: f64, bound_upper: f64, value_lower: f64, value_upper: f64) -> AndersonBjorckRootFinder {
		return AndersonBjorckRootFinder {
			bound_lower: bound_lower,
			bound_upper: bound_upper,
			value_lower: value_lower,
			value_upper: value_upper,
			last_sign: 0.0
		}
	}
	
	pub fn update(&mut self, guess: f64, value: f64) {
		if value > 0.0 {
			if self.value_lower > 0.0 || self.value_upper < 0.0 {
				if self.last_sign == 1.0 {
					// Anderson–Björck adjustment
					let m = 1.0 - value / self.value_lower;
					if m > 0.0 {
						self.value_upper *= m;
					} else {
						self.value_upper *= 0.5;
					}
				}
				
				self.bound_lower = guess;
				self.value_lower = value;
			} else {
				if self.last_sign == 1.0 {
					let m = 1.0 - value / self.value_upper;
					if m > 0.0 {
						self.value_lower *= m;
					} else {
						self.value_lower *= 0.5;
					}
				}
				
				self.bound_upper = guess;
				self.value_upper = value;
			}
			self.last_sign = 1.0;
		} else {
			if self.value_lower < 0.0 || self.value_upper > 0.0 {
				if self.last_sign == -1.0 {
					let m = 1.0 - value / self.value_lower;
					if m > 0.0 {
						self.value_upper *= m;
					} else {
						self.value_upper *= 0.5;
					}
				}
				
				self.bound_lower = guess;
				self.value_lower = value;
			} else {
				if self.last_sign == -1.0 {
					let m = 1.0 - value / self.value_upper;
					if m > 0.0 {
						self.value_lower *= m;
					} else {
						self.value_lower *= 0.5;
					}
				}
				
				self.bound_upper = guess;
				self.value_upper = value;
			}
			self.last_sign = -1.0;
		}
	}
	
	pub fn next_guess(&self) -> f64 {
		if self.value_lower.is_nan() || self.value_upper.is_nan() {
			// Fall back to interval bisection
			return (self.bound_lower + self.bound_upper) / 2.0;
		} else {
			// Regula falsi
			return (self.bound_lower * self.value_upper - self.bound_upper * self.value_lower) / (self.value_upper - self.value_lower);
		}
	}
	
	pub fn bounds(&self) -> (f64, f64) {
		return (self.bound_lower, self.bound_upper);
	}
	
	pub fn precision(&self) -> f64 {
		return (self.bound_upper - self.bound_lower).abs();
	}
}
